# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class detour."""

import contextlib
import inspect
import threading
import types
from typing import Any, Dict, Sequence, Tuple, Type, Union


@contextlib.contextmanager
def detour(
    mappings: Sequence[Tuple[
        Type[Any],                               # Source class
        Union[Type[Any], types.FunctionType]     # Target class or function
    ]]):
  """Context manager for detouring object creation.

  At times, we want to replace an object of a class to an object of a different
  class. Usually, we do so by passing the object as a function argument using
  dependency injection. However, it's not always possible to expose those
  internal objects as parameters to the class, as we cannot predict what needs
  to be customized in future. Also, exposing too many arguments will hurt
  usability, it's big burden to figure out 20 arguments of a function for a user
  to get started.

  `pg.detour` provides another option for object replacement in Python, which
  creates a context in which some source classes can be detoured to specified
  destination classes or functions. For example, the code snippet below will
  detour instantation of class A to class B, and vice-versa::

    class A:
      pass

    class B:
      pass

    # Exchange class A and class B.
    with pg.detour([(A, B), (B, A)]):
      a = A()   # a is a B object.
      b = B()   # b is an A object.

  Detour destination can be a function, which allows users to intercept the
  arguments passed to the class constructor. For example::

    class Foo:
      def __init__(self, value):
        self.value = value

    class Bar:
      def __init__(self, value):
        self.value = value

    def detoured_foo(cls, value):
      # cls is the original class before detour.
      return Bar(value + 1)

    with pg.detour([(Foo, detoured_foo)]):
      f = Foo(1)   # f will be Bar(2).

  Detour can be nested. The outer scope mappings take precedence over the
  mappings from the inner loop, allowing users to change object creation
  behaviors from the outside. For example, the following code will detour
  class A to class C::

    with pg.detour([(A, C)]):
      with pg.detour([A, B]):
        a = A()   # a is a C object.

  Detour is transisive across the inner and outer scope. For example, the code
  below will detour class A to class C through B::

    with pg.detour([(B, C)]):
      a1 = A()     # a1 is an A object.
      with pg.detour([A, B]):
        a2 = A()    # a2 is a C object. (A -> B -> C)

  Detour is thread-sfe.

  Args:
    mappings: A sequence of tuple (src_cls, dest_cls_or_fn) as mappings for the
      detour - 'src_cls' is the source class to be detoured, while
      'dest_cls_or_fn' is the destination class or function. When it's a class,
      its `__init__` method should have the same signature as the `__init__` of
      the original class. When it's a function, it should accept a positional
      argument `cls`, for passing the original class that is being detoured,
      followed by all the arguments that the original class should accept. For
      example, a class with `__init__(self, x, *args, y, **kwargs)` can be
      detoured to a function with signature `(cls, x, *args, y, **kwargs)`.

  Yields:
    Resolved detour mappings.

  Raises:
    TypeError: If the first item in each mapping is not a class, or the second
      item in each mapping is neither a class nor a function.
  """
    # Placeholder for Google-internal usage instrumentation.

  for src, dest in mappings:
    if not inspect.isclass(src):
      raise TypeError(f'Detour source {src!r} is not a class.')
    if not inspect.isclass(dest) and not inspect.isfunction(dest):
      raise TypeError(
          f'Detour destination {dest!r} is not a class or a function.')

  try:
    yield _global_detour_context.enter_scope(mappings)
  finally:
    _global_detour_context.leave_scope()


def current_mappings() -> Dict[Type[Any], Union[Type[Any], types.FunctionType]]:
  """Returns detour mappings under current scope."""
  return _global_detour_context.current_mappings


def undetoured_new(cls, *args, **kwargs) -> Any:
  """Create a new instance of cls without detouring.

  If cls.__init__ creates sub-objects, creation of sub-objects
  maybe detoured based on current context. For example::

    class A:

      def __init__(self, x):
        if x < 0:
          self.child = A(x)
        else:
          self.x = x

    with pg.detour([A, B]):
      a = A(-1)
      assert isinstance(a, A)
      assert isinstance(a.child, B)

  Args:
    cls: The class whose instance will be created.
    *args: Positional arguments to be passed to class __init__ method.
    **kwargs: Keyword arguments to be passed to class __init__ method.

  Returns:
    A instance of `cls`.
  """
  new_method = _global_detour_context.get_original_new(cls)
  if new_method is object.__new__:
    instance = new_method(cls)
  else:
    instance = new_method(cls, *args, **kwargs)
  instance.__init__(*args, **kwargs)
  return instance


class _DetourContext:
  """Context that sets/gets detoured class mappings under current thread."""

  _DETOUR_STACK_KEY = 'detour_stack'
  _DETOUR_MAPPING_KEY = 'detour_map'
  _NEW_CALL_STACK = 'new_stack'

  def __init__(self):
    self._tls = threading.local()
    self._original_new = dict()

  @property
  def _detour_stack(self):
    detour_stack = getattr(self._tls, self._DETOUR_STACK_KEY, None)
    if detour_stack is None:
      detour_stack = []
      setattr(self._tls, self._DETOUR_STACK_KEY, detour_stack)
    return detour_stack

  @property
  def current_mappings(
      self) -> Dict[Type[Any], Union[Type[Any], types.FunctionType]]:
    if self._detour_stack:
      return self._detour_stack[-1]
    return dict()

  def enter_scope(
      self,
      mappings: Sequence[Tuple[
          Type[Any],                            # Source class
          Union[Type[Any], types.FunctionType]  # Target class or function
      ]]) -> Dict[Type[Any], Union[Type[Any], types.FunctionType]]:
    """Enter a new scope."""
    # Create a copy of current mapping so we can modify.
    cur_mappings = dict(self.current_mappings)

    # Compute transisive mappings between current scope and new scope.
    # 1) If a source cls exists in current mappings, use the existing
    #    destination. E.g:
    #    ```
    #    with pg.detour([(A, B)]):
    #      with pg.detour([(A, C]]):
    #        A()   # should produce B.
    #
    # 2) If a destination cls exists in current mapping, and source class does
    #    not exist in current mappings, inserts a mapping from the source class
    #    to the destination of the target class in current mapping. E.g:
    #    ```
    #    with pg.detour([(B, C)]):
    #      with pg.detour([(A, B)]):
    #        A()   # should produce C.
    #
    # 3) Otherwise insert the new mapping. E.g:
    #    ```
    #    with pg.detour([(A, B)]):
    #      with pg.detour([(C, D)]):
    #        C()   # should produce D.
    new_mappings = []
    for src, dest in mappings:
      if src not in cur_mappings:
        if dest in cur_mappings:
          new_mappings.append((src, cur_mappings[dest]))
        else:
          new_mappings.append((src, dest))

    for src, dest in new_mappings:
      if src not in self._original_new:
        self._original_new[src] = src.__new__
        setattr(src, '__new__', _maybe_detoured_new)
      cur_mappings[src] = dest
    self._detour_stack.append(cur_mappings)
    return cur_mappings

  def leave_scope(self):
    """Leave current detour scope."""
    assert self._detour_stack
    self._detour_stack.pop(-1)

  def get_destination(self, src_cls):
    return self.current_mappings.get(src_cls, None)

  def get_original_new(self, src_cls):
    """Returns the original new method of source cls."""
    if not _is_detoured_new(src_cls.__new__):
      orig_new = src_cls.__new__
    else:
      # NOTE(daiyip): there are usually 3 patterns in implementing __new__.
      # 1) call super.__new__ to return an instance.
      # 2) explicitly call object.__new__ to return an instance.
      # 3) return an instance from another class.
      #
      # The following code aims to support case #1 by mimicing the call
      # convention of super.__new__ without access to the super object.
      # We implement this by maintaining a call history of `__new__` method
      # returned by `get_original_new` for each top-most call to
      # `_maybe_detour_new`. Based on the history, we always return the next
      # __new__ along the inheritance hierarchy. For example, for code:
      #
      # ```
      #   class A:
      #     def __new__(cls, *args, **kwargs):
      #       return super(A, cls).__new__(cls, *args, **kwargs)
      #
      #   class B:
      #     def __new__(cls, *args, **kwargs):
      #       return super(A, cls).__new__(cls, *args, **kwargs)
      #
      #   class C(A, B):
      #     pass
      # ```
      # when we detour A and B to other classes, their `__new__` method will be
      # replaced with `_maybe_detoured_new`. As we create an object of C, it
      # will call `C.__new__`, which inherits the `_maybe_detoured_new` assigned
      # to `A.__new__`. `_maybe_detoured_new` calls `get_original_new` on class
      # C, which should return the original `A.__new__`. It then executes
      # `super(A, cls).__new__`, which triggers `_maybe_detoured_new` method
      # again assigned to `B.__new__`. In such case, we cannot differentiate the
      # first call to `_maybe_detoured_new` (C.__new__) from this call, since
      # both take class C as the cls argument. However, by assuming that nested
      # `_maybe_detoured_new` call should always reflect the `super.__new__`
      # call convention, we can store the call history for these invoked __new__
      # methods, and return the one that is one-step closer to `object.__new__`.
      # This may not work for the most complicated __new__ customization, but
      # should work well for most __new__ implementations.
      orig_new = self._original_new.get(src_cls, object.__new__)
      if orig_new is object.__new__ or orig_new in self._new_stack:
        for base in src_cls.__bases__:
          base_new = self.get_original_new(base)
          if base_new is not object.__new__ and base_new not in self._new_stack:
            orig_new = base_new
            break
    return orig_new

  @property
  def _new_stack(self):
    """Returns the stack of new methods in current thread."""
    stack = getattr(self._tls, self._NEW_CALL_STACK, None)
    if stack is None:
      stack = []
      setattr(self._tls, self._NEW_CALL_STACK, stack)
    return stack

  def call_new(self, new_method, cls, *args, **kwargs):
    """Call __new__ method with correctly handling super.__new__."""
    try:
      self._new_stack.append(new_method)
      if new_method is object.__new__:
        return object.__new__(cls)
      else:
        return new_method(cls, *args, **kwargs)
    finally:
      self._new_stack.pop(-1)

# Global detour context.
_global_detour_context = _DetourContext()


@staticmethod   # This decorator is required for Python2.
def _maybe_detoured_new(cls, *args, **kwargs):
  """A __new__ method to replace user class' __new__ for detour."""
  dest_cls_or_fn = _global_detour_context.get_destination(cls)
  if dest_cls_or_fn is None:
    # No detour in current thread.
    return _global_detour_context.call_new(
        _global_detour_context.get_original_new(cls),
        cls, *args, **kwargs)

  if inspect.isclass(dest_cls_or_fn):
    dest_cls = dest_cls_or_fn
    instance = _global_detour_context.call_new(
        _global_detour_context.get_original_new(dest_cls),
        dest_cls, *args, **kwargs)

    # NOTE(daiyip): when an overridden `__new__` returns an instance whose
    # class is not strictly the user class, `__init__` will not be called
    # by Python runtime. We can detect such case and invoke `__init__`
    # manually.
    if not isinstance(instance, cls):
      instance.__init__(*args, **kwargs)
    return instance
  else:
    # NOTE(daiyip): when function is used as detour destination, we handle it
    # specially to allow instances of the source class to be created in the
    # function. E.g.
    #
    # ```
    # def create_foo(value):
    #   return Foo(value + 1)
    #
    # with pg.detour([(Foo, create_foo)]):
    #    Foo(1)
    # ```
    try:
      _global_detour_context.current_mappings[cls] = cls
      return dest_cls_or_fn(cls, *args, **kwargs)
    finally:
      _global_detour_context.current_mappings[cls] = dest_cls_or_fn


def _is_detoured_new(method):
  """Returns if a method is detoured new."""
  return method is getattr(_maybe_detoured_new, '__func__')



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pluggable record IO."""

import abc
import collections
import os
import re
from typing import Any, Callable, Iterator, Optional, Union

from pyglove.core.io import file_system


class Sequence(metaclass=abc.ABCMeta):
  """Interface for a sequence of records."""

  def __init__(
      self,
      perms: Optional[int] = None,
      serializer: Optional[Callable[[Any], Union[bytes, str]]] = None,
      deserializer: Optional[Callable[[Union[bytes, str]], Any]] = None
  ):
    self._perms = perms
    self._serializer = serializer
    self._deserializer = deserializer

  def add(self, record: Any) -> None:
    """Adds a record to the reader."""
    if self._serializer:
      record = self._serializer(record)
    if not isinstance(record, (str, bytes)):
      raise ValueError(
          f'Cannot write record with type {type(record)}. '
          'Did you forget to pass a serializer?'
      )
    self._add(record)

  @abc.abstractmethod
  def _add(self, record: Union[bytes, str]) -> None:
    """Adds a raw record to the reader."""

  @abc.abstractmethod
  def __len__(self):
    """Gets the number of records in the reader."""

  def __iter__(self) -> Iterator[Any]:
    """Iterates over the records in the reader."""
    for record in self._iter():
      if self._deserializer:
        yield self._deserializer(record)
      else:
        yield record

  @abc.abstractmethod
  def _iter(self) -> Iterator[Union[bytes, str]]:
    """Iterates over the raw records in the reader."""

  @abc.abstractmethod
  def close(self) -> None:
    """Closes the reader."""

  @abc.abstractmethod
  def flush(self) -> None:
    """Flushes the read records to the storage."""

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    del exc_type, exc_value, traceback
    self.close()


class SequenceIO(metaclass=abc.ABCMeta):
  """Interface for a record IO system."""

  @abc.abstractmethod
  def open(
      self,
      path: Union[str, os.PathLike[str]],
      mode: str,
      *,
      perms: Optional[int],
      serializer: Optional[Callable[[Any], Union[bytes, str]]],
      deserializer: Optional[Callable[[Union[bytes, str]], Any]],
      **kwargs
  ) -> Sequence:
    """Opens a sequence for reading or writing."""


_SHARDED_FILE_EXT = re.compile(r'(.*)-(\d+)-of-(\d+)$')


class _SequenceIORegistry(object):
  """Registry for record IO systems."""

  def __init__(self):
    self._registry = {}

  def add(self, extension: str, sequence_io: SequenceIO) -> None:
    """Adds a record IO system with a prefix."""
    self._registry[extension] = sequence_io

  def get(self, path: Union[str, os.PathLike[str]]) -> SequenceIO:
    """Gets the record IO system for a path."""
    path = file_system.resolve_path(path)
    parts = path.split('.')
    if parts:
      extension = parts[-1].lower()
      if '@' in extension:
        extension = extension.split('@')[0]
      match = _SHARDED_FILE_EXT.match(extension)
      if match:
        extension = match.group(1)
      if extension in self._registry:
        return self._registry[extension]
    return LineSequenceIO()


_registry = _SequenceIORegistry()


def add_sequence_io(extension: str, sequence_io: SequenceIO) -> None:
  """Adds a record IO system with a prefix."""
  _registry.add(extension, sequence_io)


def open_sequence(
    path: Union[str, os.PathLike[str]],
    mode: str = 'r',
    *,
    perms: Optional[int] = 0o664,  # Default to world-readable.
    serializer: Optional[
        Callable[[Any], Union[bytes, str]]
    ] = None,
    deserializer: Optional[
        Callable[[Union[bytes, str]], Any]
    ] = None,
    make_dirs_if_not_exist: bool = True,
) -> Sequence:
  """Open sequence for reading or writing.

  Args:
    path: The path to the sequence.
    mode: The mode of the sequence.
    perms: (Optional) The permissions of the sequence.
    serializer: (Optional) A serializer function for converting a structured
      object to a string or bytes.
    deserializer: (Optional) A deserializer function for converting a string or
      bytes to a structured object.
    make_dirs_if_not_exist: (Optional) Whether to create the directories
      if they do not exist. Applicable when opening in write or append mode.

  Returns:
    A sequence for reading or writing.
  """
  if 'w' in mode or 'a' in mode:
    parent_dir = os.path.dirname(path)
    if make_dirs_if_not_exist:
      file_system.mkdirs(parent_dir, exist_ok=True)
  return _registry.get(path).open(
      path, mode, perms=perms, serializer=serializer, deserializer=deserializer
  )


class MemorySequence(Sequence):
  """An in-memory sequence."""

  def __init__(
      self,
      path: str,
      mode: str,
      records: list[Union[str, bytes]],
      *,
      perms: Optional[int],
      serializer: Optional[Callable[[Any], Union[bytes, str]]],
      deserializer: Optional[Callable[[Union[bytes, str]], Any]]
  ):
    super().__init__(perms, serializer, deserializer)
    self._path = path
    self._mode = mode
    self._records = records
    self._closed = False

  def _add(self, record: Union[str, bytes]) -> None:
    if 'w' not in self._mode and 'a' not in self._mode:
      raise ValueError(
          f'Cannot write record {record!r} to memory sequence {self._path!r} '
          f'with mode {self._mode!r}.'
      )
    if self._closed:
      raise ValueError(
          f'Cannot write record {record!r} to a closed writer for '
          f'{self._path!r}.'
      )
    self._records.append(record)

  def __len__(self):
    return len(self._records)

  def _iter(self):
    if 'r' not in self._mode:
      raise ValueError(
          f'Cannot read memory sequence {self._path!r} with '
          f'mode {self._mode!r}.'
      )
    if self._closed:
      raise ValueError(
          f'Cannot iterate over a closed sequence reader {self._path!r}.'
      )
    return iter(self._records)

  def flush(self):
    pass

  def close(self) -> None:
    self._closed = True


class MemorySequenceIO(SequenceIO):
  """Memory-based record IO."""

  def __init__(self):
    super().__init__()
    self._root = collections.defaultdict(list)

  def open(
      self,
      path: Union[str, os.PathLike[str]],
      mode: str,
      *,
      perms: Optional[int],
      serializer: Optional[Callable[[Any], Union[bytes, str]]],
      deserializer: Optional[Callable[[Union[bytes, str]], Any]],
      **kwargs
  ) -> Sequence:
    """Opens a reader for a sequence."""
    del kwargs
    if 'w' in mode:
      self._root[path] = []
    return MemorySequence(
        path, mode, self._root[path],
        perms=perms, serializer=serializer, deserializer=deserializer
    )


add_sequence_io('mem', MemorySequenceIO())


class LineSequence(Sequence):
  """A new-line broken sequence."""

  def __init__(
      self,
      path: str,
      mode: str,
      perms: Optional[int],
      serializer: Optional[Callable[[Any], Union[bytes, str]]],
      deserializer: Optional[Callable[[Union[bytes, str]], Any]],
  ) -> None:
    super().__init__(perms, serializer, deserializer)
    self._path = path
    self._mode = mode
    self._file = file_system.open(path, mode)

  def __len__(self):
    raise NotImplementedError(
        '__len__ is not supported for LineSequence. '
        'Use `len(list(iter(sequence)))` instead.'
    )

  def _iter(self):
    while True:
      line = self._file.readline()
      if not line:
        break
      yield line.rstrip('\n')

  def _add(self, record: Union[str, bytes]) -> None:
    self._file.write(record.rstrip('\n'))
    self._file.write('\n')

  def flush(self) -> None:
    self._file.flush()

  def close(self) -> None:
    self._file.close()
    if ('w' in self._mode or 'a' in self._mode) and self._perms is not None:
      file_system.chmod(self._path, self._perms)


class LineSequenceIO(SequenceIO):
  """Line-based record IO."""

  def open(
      self,
      path: Union[str, os.PathLike[str]],
      mode: str,
      *,
      perms: Optional[int],
      serializer: Optional[Callable[[Any], Union[bytes, str]]],
      deserializer: Optional[Callable[[Union[bytes, str]], Any]],
      **kwargs
  ) -> Sequence:
    """Opens a reader for a sequence."""
    del kwargs
    return LineSequence(path, mode, perms, serializer, deserializer)




# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import unittest
from pyglove.core.io import sequence as sequence_io
# We need to import this module to register the default parser/serializer.
import pyglove.core.symbolic as pg_symbolic


class LineSequenceIOTest(unittest.TestCase):

  def test_read_write(self):
    tmp_dir = tempfile.gettempdir()
    file1 = os.path.join(tmp_dir, 'abc', 'file1')
    with pg_symbolic.open_jsonl(file1, 'w') as f:
      self.assertIsInstance(f, sequence_io.LineSequence)
      f.add(1)
      f.add(' foo')
      f.add(' bar ')
      f.flush()
      self.assertTrue(os.path.exists(file1))

      f.add('baz\n')
      f.add(dict(x=1))

    self.assertTrue(os.path.exists(file1))
    with pg_symbolic.open_jsonl(file1, 'r') as f:
      self.assertIsInstance(f, sequence_io.LineSequence)
      self.assertEqual(list(iter(f)), [1, ' foo', ' bar ', 'baz\n', dict(x=1)])

    with pg_symbolic.open_jsonl(file1, 'a') as f:
      f.add('qux')

    with pg_symbolic.open_jsonl(file1, 'r') as f:
      with self.assertRaisesRegex(
          NotImplementedError, '__len__ is not supported'
      ):
        _ = len(f)
      self.assertEqual(
          list(iter(f)), [1, ' foo', ' bar ', 'baz\n', dict(x=1), 'qux']
      )

  def test_read_write_with_raw_texts(self):
    tmp_dir = tempfile.gettempdir()
    file2 = os.path.join(tmp_dir, 'file2')
    with sequence_io.open_sequence(file2, 'w') as f:
      self.assertIsInstance(f, sequence_io.LineSequence)
      with self.assertRaisesRegex(
          ValueError, 'Cannot write record with type'
      ):
        f.add(1)
      f.add('foo\nbar\n')

    with sequence_io.open_sequence(file2, 'r') as f:
      with self.assertRaisesRegex(
          ValueError, 'not writable'
      ):
        f.add('baz')
      self.assertEqual(list(iter(f)), ['foo', 'bar'])


class MemorySequenceIOTest(unittest.TestCase):

  def test_read_write(self):
    with sequence_io.open_sequence('/file1.mem@123', 'w') as f:
      self.assertIsInstance(f, sequence_io.MemorySequence)
      f.add(' foo')
      f.add(' bar ')
      f.flush()
      f.add('baz')
      with self.assertRaisesRegex(
          ValueError, 'Cannot write record with type'
      ):
        f.add(1)
      with self.assertRaisesRegex(
          ValueError, 'Cannot read memory sequence'
      ):
        next(iter(f))

    with self.assertRaisesRegex(
        ValueError, 'Cannot write record .* to a closed writer'
    ):
      f.add('qux')

    with sequence_io.open_sequence('/file1.mem@123', 'a') as f:
      self.assertIsInstance(f, sequence_io.MemorySequence)
      f.add('qux')

    with sequence_io.open_sequence('/file1.mem@123') as f:
      self.assertIsInstance(f, sequence_io.MemorySequence)
      self.assertEqual(len(f), 4)
      self.assertEqual(list(f), [' foo', ' bar ', 'baz', 'qux'])
      with self.assertRaisesRegex(
          ValueError, 'Cannot write record .* to memory sequence'
      ):
        f.add('abc')

    with self.assertRaisesRegex(
        ValueError, 'Cannot iterate over a closed sequence reader'
    ):
      next(iter(f))

    with sequence_io.open_sequence('/file1.mem@123', 'w') as f:
      f.add('abc')

    with sequence_io.open_sequence('/file1.mem@123', 'r') as f:
      self.assertEqual(list(iter(f)), ['abc'])

  def test_sharded_file_name(self):
    with sequence_io.open_sequence('/file1.mem-00000-of-00001', 'w') as f:
      self.assertIsInstance(f, sequence_io.MemorySequence)


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Numerical hyper primitives."""

import typing
from typing import Any, Callable, Optional, Tuple

from pyglove.core import geno
from pyglove.core import symbolic
from pyglove.core import typing as pg_typing
from pyglove.core import utils
from pyglove.core.hyper import base


@symbolic.members(
    [
        ('min_value', pg_typing.Float(), 'Minimum acceptable value.'),
        ('max_value', pg_typing.Float(), 'Maximum acceptable value.'),
        geno.float_scale_spec('scale'),
    ],
    init_arg_list=['min_value', 'max_value', 'scale', 'name', 'hints'],
    serialization_key='hyper.Float',
    additional_keys=['pyglove.generators.genetic.Float']
)
class Float(base.HyperPrimitive):
  """A continuous value within a range.

  Example::

    # A float value between between 0.0 and 1.0.
    v = pg.floatv(0.0, 1.0)

  See also:

    * :func:`pyglove.floatv`
    * :class:`pyglove.hyper.Choices`
    * :class:`pyglove.hyper.OneOf`
    * :class:`pyglove.hyper.ManyOf`
    * :class:`pyglove.hyper.CustomHyper`
  """

  def _on_bound(self):
    """Constructor."""
    super()._on_bound()
    if self.min_value > self.max_value:
      raise ValueError(
          f'\'min_value\' ({self.min_value}) is greater than \'max_value\' '
          f'({self.max_value}).')
    if self.scale in ['log', 'rlog'] and self.min_value <= 0:
      raise ValueError(
          f'\'min_value\' must be positive when `scale` is {self.scale!r}. '
          f'encountered: {self.min_value}.')

  def dna_spec(self, location: Optional[utils.KeyPath] = None) -> geno.Float:
    """Returns corresponding DNASpec."""
    return geno.Float(
        min_value=self.min_value,
        max_value=self.max_value,
        scale=self.scale,
        hints=self.hints,
        name=self.name,
        location=location or utils.KeyPath(),
    )

  def _decode(self) -> float:
    """Decode a DNA into a float value."""
    dna = self._dna
    if not isinstance(dna.value, float):
      raise ValueError(
          utils.message_on_path(
              f'Expect float value. Encountered: {dna.value}.', self.sym_path
          )
      )
    if dna.value < self.min_value:
      raise ValueError(
          utils.message_on_path(
              f'DNA value should be no less than {self.min_value}. '
              f'Encountered {dna.value}.',
              self.sym_path,
          )
      )

    if dna.value > self.max_value:
      raise ValueError(
          utils.message_on_path(
              f'DNA value should be no greater than {self.max_value}. '
              f'Encountered {dna.value}.',
              self.sym_path,
          )
      )
    return dna.value

  def encode(self, value: float) -> geno.DNA:
    """Encode a float value into a DNA."""
    if not isinstance(value, float):
      raise ValueError(
          utils.message_on_path(
              f'Value should be float to be encoded for {self!r}. '
              f'Encountered {value}.',
              self.sym_path,
          )
      )
    if value < self.min_value:
      raise ValueError(
          utils.message_on_path(
              f'Value should be no less than {self.min_value}. '
              f'Encountered {value}.',
              self.sym_path,
          )
      )
    if value > self.max_value:
      raise ValueError(
          utils.message_on_path(
              f'Value should be no greater than {self.max_value}. '
              f'Encountered {value}.',
              self.sym_path,
          )
      )
    return geno.DNA(value)

  def custom_apply(
      self,
      path: utils.KeyPath,
      value_spec: pg_typing.ValueSpec,
      allow_partial: bool = False,
      child_transform: Optional[
          Callable[[utils.KeyPath, pg_typing.Field, Any], Any]
      ] = None,
  ) -> Tuple[bool, 'Float']:
    """Validate candidates during value_spec binding time."""
    del allow_partial
    del child_transform
    # Check if value_spec directly accepts `self`.
    if value_spec.value_type and isinstance(self, value_spec.value_type):
      return (False, self)

    float_spec = typing.cast(
        pg_typing.Float, pg_typing.ensure_value_spec(
            value_spec, pg_typing.Float(), path))
    if float_spec:
      if (float_spec.min_value is not None
          and self.min_value < float_spec.min_value):
        raise ValueError(
            utils.message_on_path(
                f'Float.min_value ({self.min_value}) should be no less than '
                f'the min value ({float_spec.min_value}) of value spec: '
                f'{float_spec}.',
                path,
            )
        )
      if (float_spec.max_value is not None
          and self.max_value > float_spec.max_value):
        raise ValueError(
            utils.message_on_path(
                f'Float.max_value ({self.max_value}) should be no greater than '
                f'the max value ({float_spec.max_value}) of value spec: '
                f'{float_spec}.',
                path,
            )
        )
    return (False, self)

  def is_leaf(self) -> bool:
    """Returns whether this is a leaf node."""
    return True


def floatv(min_value: float,
           max_value: float,
           scale: Optional[str] = None,
           *,
           name: Optional[str] = None,
           hints: Optional[Any] = None) -> Any:
  """A continuous value within a range.

  Example::

    # A continuous value within [0.0, 1.0]
    v = pg.floatv(0.0, 1.0)

  See also:

    * :class:`pyglove.hyper.Float`
    * :func:`pyglove.oneof`
    * :func:`pyglove.manyof`
    * :func:`pyglove.permutate`
    * :func:`pyglove.evolve`

  .. note::

    Under symbolic mode (by default), `pg.floatv` returns a ``pg.hyper.Float``
    object. Under dynamic evaluate mode, which is called under the context of
    :meth:`pyglove.hyper.DynamicEvaluationContext.collect` or
    :meth:`pyglove.hyper.DynamicEvaluationContext.apply`, it evaluates to
    a concrete candidate value.

  Args:
    min_value: Minimum acceptable value (inclusive).
    max_value: Maximum acceptable value (inclusive).
    scale: An optional string as the scale of the range. Supported values
      are None, 'linear', 'log', and 'rlog'.
      If None, the feasible space is unscaled.
      If `linear`, the feasible space is mapped to [0, 1] linearly.
      If `log`, the feasible space is mapped to [0, 1] logarithmically with
        formula `x -> log(x / min) / log(max / min)`.
      If `rlog`, the feasible space is mapped to [0, 1] "reverse"
        logarithmically, resulting in values close to `max_value` spread
        out more than the points near the `min_value`, with formula:
        x -> 1.0 - log((max + min - x) / min) / log (max / min).
      `min_value` must be positive if `scale` is not None.
      Also, it depends on the search algorithm to decide whether this
      information is used or not.
    name: A name that can be used to identify a decision point in the search
      space. This is needed when the code to instantiate the same hyper
      primitive may be called multiple times under a
      `pg.DynamicEvaluationContext.collect` context or a
      `pg.DynamicEvaluationContext.apply` context.
    hints: An optional value which acts as a hint for the controller.

  Returns:
    In symbolic mode, this function returns a `Float`.
    In dynamic evaluate mode, this function returns a float value that is no
    less than the `min_value` and no greater than the `max_value`.
    If evaluated under an `pg.DynamicEvaluationContext.apply` scope,
    this function will return a chosen float value from the controller
    decisions.
    If evaluated under a `pg.DynamicEvaluationContext.collect`
    scope, it will return `min_value`.
  """
  return Float(
      min_value=min_value, max_value=max_value,
      scale=scale, name=name, hints=hints)



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import concurrent.futures
import unittest
from pyglove.core.utils import contextual


class ContextualTest(unittest.TestCase):

  def test_contextual_override(self):
    with contextual.contextual_override(x=3, y=3, z=3) as parent_override:
      self.assertEqual(
          parent_override,
          dict(
              x=contextual.ContextualOverride(
                  3, cascade=False, override_attrs=False
              ),
              y=contextual.ContextualOverride(
                  3, cascade=False, override_attrs=False
              ),
              z=contextual.ContextualOverride(
                  3, cascade=False, override_attrs=False
              ),
          ),
      )
      self.assertEqual(
          contextual.get_contextual_override('y'),
          contextual.ContextualOverride(3, cascade=False, override_attrs=False),
      )
      self.assertEqual(contextual.contextual_value('x'), 3)
      self.assertIsNone(contextual.contextual_value('f', None))
      with self.assertRaisesRegex(KeyError, '.* does not exist'):
        contextual.contextual_value('f')

      self.assertEqual(contextual.all_contextual_values(), dict(x=3, y=3, z=3))

      # Test nested contextual override with override_attrs=True (default).
      with contextual.contextual_override(
          y=4, z=4, override_attrs=True) as nested_override:
        self.assertEqual(
            nested_override,
            dict(
                x=contextual.ContextualOverride(
                    3, cascade=False, override_attrs=False
                ),
                y=contextual.ContextualOverride(
                    4, cascade=False, override_attrs=True
                ),
                z=contextual.ContextualOverride(
                    4, cascade=False, override_attrs=True
                ),
            ),
        )

    # Test nested contextual override with cascade=True.
    with contextual.contextual_override(x=3, y=3, z=3, cascade=True):
      with contextual.contextual_override(y=4, z=4, cascade=True):
        self.assertEqual(contextual.contextual_value('x'), 3)
        self.assertEqual(contextual.contextual_value('y'), 3)
        self.assertEqual(contextual.contextual_value('z'), 3)

  def test_with_contextual_override(self):
    def func(i):
      del i
      return contextual.contextual_value('x')

    pool = concurrent.futures.ThreadPoolExecutor()
    with contextual.contextual_override(x=3):
      self.assertEqual(contextual.with_contextual_override(func)(0), 3)
      self.assertEqual(
          list(pool.map(contextual.with_contextual_override(func), range(1))),
          [3]
      )

  def test_with_contextual_override_async_func(self):
    async def func(i):
      del i
      return contextual.contextual_value('x')

    with contextual.contextual_override(x=3):
      self.assertEqual(
          asyncio.run(contextual.with_contextual_override(func)(0)), 3
      )

if __name__ == '__main__':
  unittest.main()
